/*
 * Mailfront Plugin Name: plugin-check-dnsbl
 *
 * This plugin used to provide DNSBL/DNSWL checking of sender's IP address and
 * to defer, reject or accept mails based on its check.  This plugin will add
 * its checking results to each mail message header for all mails accepted
 * depending on environment variables configurations.
 *
 * (1).  If $DNSBL_CHECK_SKIP is set, this plugin will be skipped.
 *
 * (2).  If $RELAYCLIENT is set, add to message header and skip the remaining
 *       checks.  Example of message header added:
 *
 *       X-DNSBL-Check-Result: SKIPPED: $RELAYCLIENT is set for 1.2.3.4
 *
 * (3).  If sender is authenticated, add to message header and skip the
 *       remaining checks.  Example of message header added:
 *
 *       X-DNSBL-Check-Result: SKIPPED: authenticated user from 1.2.3.4
 *
 * (4).  If $PROTO is TCP6 or ${PROTO}REMOTEIP contain ":" character, add to
 *       message header and skip the remaining checks. Example of message
 *       header added:
 *
 *       X-DNSBL-Check-Result: SKIPPED: $PROTO is TCP6 or is non-IPv4 for ${PROTO}REMOTEIP
 *
 *       ${PROTO}REMOTEIP will be parsed.
 *
 * (5).  If $RBLSMTPD is set and empty, add to message header and skip the
 *       remaining checks.  If $RBLSMTPD is set and not empty, the reply code
 *       is set to 553 if the first character of $RBLSMTPD is "-" otherwise
 *       the reply code is set to 451.  The reply message will be copied from
 *       $RBLSMTPD and will skip "-" if it is the starting character of
 *       $RBLSMTPD.  Examples of smtp.rules for tcprules:
 *
 *       1.1.1.1:allow,RBLSMTPD=""
 *       2.2.2.2:allow,RBLSMTPD="-Your IP is listed in our denied list."
 *       3.3.3.3:allow,RBLSMTPD="Your IP is listed in our denied list."
 *
 *       The above will cause this plugin to accept mails from IP 1.1.1.1,
 *       reject mails from IP 2.2.2.2 and defer mails from IP 3.3.3.3.  Below
 *       are the reply to sender's IP 2.2.2.2 and 3.3.3.3 respectively:
 *
 *       553 Your IP is listed in our denied list.
 *       451 Your IP is listed in our denied list.
 *
 *       Example of message header added for IP 1.1.1.1:
 *
 *       X-DNSBL-Check-Result: SKIPPED: $RBLSMTPD is set and empty for 1.1.1.1
 *
 * (6).  $RWLDOMAINS is the list of servers/domains to check for whitelisting
 *       (DNSWL).  If $RWLDOMAINS is set and not empty, the domains listed
 *       will be used for DNSWL checking.  Each domain needs to be seperated
 *       with a SPACE or a colon ":".
 *
 * (7).  $RWLDOMAINSFILE is a file with list of servers/domains to check for
 *       whitelisting (DNSWL).   This will only be used if $RWLDOMAINS is not
 *       set.  i.e. copy the content into $RWLDOMAINS.
 *       If $RWLDOMAINSFILE not set will default to
 *       /var/qmail/control/rwldomains which is a define constant of
 *       DNSBL_RWLDOMAINS_FILE.  Format in this file is one domain per line.
 *       Lines started with # are ignored.
 *
 * (8).  $RBLDOMAINS is the list of servers/domains to check for blacklisting
 *       (DNSBL).  If $RBLDOMAINS is set and not empty, the domains listed
 *       will be used for DNSBL checking.  Each domain needs to be seperated
 *       with a SPACE or a colon ":".
 *
 * (9).  $RBLDOMAINSFILE is a file with list of servers/domains to check for
 *       blacklisting (DNSBL).   This will only be used if $RBLDOMAINS is not
 *       set.  i.e. copy the content into $RBLDOMAINS.
 *       If $RBLDOMAINSFILE not set will default to
 *       /var/qmail/control/rbldomains which is a define constant of
 *       DNSBL_RBLDOMAINS_FILE.  Format in this file is one domain per line.
 *       Lines started with # are ignored.
 *
 * (10). If $RWLDOMAINS and $RBLDOMAINS not set or empty, add to mail
 *       message header and skip the DNSWL/DNSBL checks.  Example of mail
 *       header added:
 *
 *       X-DNSBL-Check-Result: SKIPPED: $RWLDOMAINS and $RBLDOMAINS not set or empty for 1.2.3.4
 *
 * (11). If $DNSBL_CHECK_REJECT is set, sender will be rejected with reply
 *       code 553 if sender's IP is listed in one of the $RBLDOMAINS domain.
 *       If sender's IP is listed in one of the $RWLDOMAINS domains, mail
 *       message will be accepted as this plugin take preference of
 *       $RWLDOMAINS domains then $RBLDOMAINS domains.
 *
 * (12). If $DNSBL_CHECK_DEFER is set, sender will be rejected with reply
 *       code 451 if sender's IP is listed in one of the $RBLDOMAINS domain.
 *       If sender's IP is listed in one of the $RWLDOMAINS domains, mail
 *       message will be accepted as this plugin take preference of
 *       $RWLDOMAINS domains then $RBLDOMAINS domains.
 *       NOTE: $DNSBL_CHECK_REJECT take preference of $DNSBL_CHECK_DEFER.
 *
 * NOTE: For (11) and (12), mails that are rejected/deferred due to blacklist,
 *       this plugin will use the first listed domain's TXT record in its
 *       reply message.  If there isn't such will default to the following:
 *
 *       Sorry, your ip ${PROTO}REMOTEIP is listed in one or more of the DNSBL: $RBLDOMAINS
 *
 *       The ${PROTO}REMOTEIP and $RBLDOMAINS will be parsed accordingly.
 *
 *       If sender's IP is listed in one of the $RWLDOMAINS domain, instead of
 *       rejecting/deferring mail, mail message will be accepted and all the
 *       DNSWL/DNSBL check results will be added to each mail message header.
 *
 * (13). If $DNSBL_CHECK_REJECT and $DNSBL_CHECK_DEFER not set, all mail
 *       message will be accepted with DNSWL/DNSBL check results added to each
 *       mail message header.
 *
 * (14). If $DNSBL_CHECK_VERBOSE is set, this plugin will log its checking to
 *       stderr.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <resolv.h>
#include <errno.h>
#include <libowfat/stralloc.h>
#include <libowfat/byte.h>
#include <libowfat/openreadclose.h>

#include <msg/msg.h>
#include "mailfront.h"

#define eprintf(...) fprintf(stderr, __VA_ARGS__)
#define elogerr(N, M, ...) { if(N){ if(M){ eprintf_logname(); } fprintf(stderr, __VA_ARGS__); } }

#ifndef MAX_DNSTXT_RECORDS
#define MAX_DNSTXT_RECORDS 10
#endif
#ifndef DNS_MAX_DNSBL_LENGTH
#define DNS_MAX_DNSBL_LENGTH 512
#endif
#ifndef DNSBL_RWLDOMAINS_FILE
#define DNSBL_RWLDOMAINS_FILE "/var/qmail/control/rwldomains"
#endif
#ifndef DNSBL_RBLDOMAINS_FILE
#define DNSBL_RBLDOMAINS_FILE "/var/qmail/control/rbldomains"
#endif

static response resp;

static const char *mainprog = "mailfront";
static const char *thisprog = "plugin-check-dnsbl";
static unsigned long thispid;
static const char *linkproto;
static int is_ipv6 = 0;
static const char *local_ip;
static const char *remote_ip;
static const char *relayclient;
static const char *rbldomains;
static const char *rwldomains;
static const char *rbldomainsfile;
static const char *rwldomainsfile;
static const char *skipcheck;
static const char *rblsmtpd;
static const char *defercheck;
static const char *rejectcheck;
static const char *logcheck;
static str headermsg;

void eprintf_logname()
{
	if (!thispid) {
		thispid = (unsigned long)getpid();
	}
	eprintf("%s-%s[%lu]: ", mainprog, thisprog, thispid);
}

/* similiar to int openreadclose(const char *filename, stralloc* sa,size_t bufsize); */
/* with a difference... set errno=ENOENT and return -1 instead of return 0 */
int openreadclose_to_sa(const char *filename, stralloc *sa, size_t bufsize)
{
	int rc = openreadclose(filename, sa, bufsize);
	if (rc == 0) {
		errno = ENOENT;
		return -1;
	} else if (rc < 0) {
		return -1;
	} else {
		return 1;
	}
}

struct HostDNSTXT {
	char *txtname;
	int size;
	char *txtrecord[MAX_DNSTXT_RECORDS];
};

void init_HostDNSTXT(struct HostDNSTXT *d)
{
	d->txtname = NULL;
	d->size = 0;
}

void free_HostDNSTXT(struct HostDNSTXT *d)
{
	int x = 0;
	if (d->size > 0) {
		for (x = 0; x < d->size; x++) {
			free(d->txtrecord[x]);
		}
		d->size = 0;
	}
	init_HostDNSTXT(d);
}

/* return 0 for success otherwise return errno=ENOMEM if out of memory */
int check_dns_txt(const char *name, struct HostDNSTXT *s, int limit)
{
	unsigned char *ansbuf = (unsigned char *)malloc(NS_PACKETSZ);
	if (ansbuf == NULL) {
		errno = ENOMEM;
		return errno;
	}

	int l, i, c;
	c = 0;
	
	l = res_query(name, C_IN, T_TXT, ansbuf, NS_PACKETSZ);
	if (l > 0 && ((HEADER *)ansbuf)->ancount != 0) {
		unsigned char *recbuf = ansbuf+NS_HFIXEDSZ;
		for (i = ns_get16(ansbuf+4); i != 0; --i) {
			recbuf += dn_skipname(recbuf, ansbuf+l)+4;
		}
		for (i = ns_get16(ansbuf+6); i != 0; --i) {

			recbuf += dn_skipname(recbuf, ansbuf+l);

			if(ns_get16(recbuf) != T_TXT) {
				recbuf += 10 + ns_get16(recbuf+8);
				continue;
			}
			/* it's a TXT record, wow */
			if (c < limit) {
				if (s->txtname == NULL) {
					s->txtname = (char *)name;
				}
				char *tmp = (char *)malloc(256);
				if (tmp == NULL) {
					errno = ENOMEM;
					free(ansbuf);
					return errno;
				}
				snprintf(tmp, recbuf[10]+1, "%s",
					 (char *)recbuf+11);
				s->txtrecord[c] = strdup(tmp);
				if (s->txtrecord[c] == NULL) {
					errno = ENOMEM;
					free(ansbuf);
					free(tmp);
					return errno;
				}
				free(tmp);
				c++;
			}
		}
	}
	free(ansbuf);
	s->size = c;
	return 0;
}

/* return the ipv4 format string suitable for dnsbl checking
 * if success will return the required format otherwise original
 * input string will be returned
 * If out of memory, it will return NULL and errno set to ENOMEM
 * NOTE: remember to free the return
 */
char *ipv4_reverse(char *s)
{
	unsigned int n1, n2, n3, n4;

	char *buf = malloc(strlen(s)+1);
	if (buf == NULL) {
		errno = ENOMEM;
		return NULL;
	}
	strcpy(buf, s);
	if (sscanf(s, "%3u.%3u.%3u.%3u", &n1, &n2, &n3, &n4) != 4) return buf;

	if ((n1 != 0) && (n1 <= 255) && (n2 <= 255) && (n3 <= 255)
	    && (n4 <= 255)) {
		sprintf(buf, "%u.%u.%u.%u", n4, n3, n2, n1);
	}
	return buf;
}

/* return 1 if is valid IPv4 format otherwise return 0 */
int is_valid_ipv4(const char *s)
{
	unsigned int n1, n2, n3, n4;

	if (sscanf(s, "%3u.%3u.%3u.%3u", &n1, &n2, &n3, &n4) != 4) return 0;

	if ((n1 != 0) && (n1 <= 255) && (n2 <= 255) && (n3 <= 255)
	    && (n4 <= 255)) {
		char buf[20];
		sprintf(buf, "%u.%u.%u.%u", n1, n2, n3, n4);
		if(strcmp(buf, s)) return 0;
		return 1;
	}
	return 0;
}

/* return 0 on success otherwise return errno=ENOMEM */
int read_dnsbwlfiles()
{
	if (!rwldomains && rwldomainsfile && *rwldomainsfile) {
		stralloc sa0;
		stralloc_init(&sa0);
		if (!stralloc_copys(&sa0, "")) {
			errno = ENOMEM;
			return errno;
		}
		size_t bufsize = 64;
		if (openreadclose_to_sa(rwldomainsfile, &sa0, bufsize) == 1) {
			if (sa0.len > 0 && strlen(sa0.s) > 0) {
				char *tmp = (char *)malloc(sa0.len+2);
				if (tmp == NULL) {
					errno = ENOMEM;
					str_init(&headermsg);
					stralloc_free(&sa0);
					return errno;
				}
				stralloc sa1;
				stralloc_init(&sa1);
				if (!stralloc_copys(&sa1, "")) {
					errno = ENOMEM;
					stralloc_free(&sa0);
					free(tmp);
					return errno;
				}
				byte_copy(tmp, sa0.len, sa0.s);
				char *pch;
				char *tmppch = strdup(tmp);
				if (tmppch == NULL) {
					errno = ENOMEM;
					stralloc_free(&sa0);
					stralloc_free(&sa1);
					free(tmp);
					return errno;
				}
				pch = strtok(tmppch, "\n");
				while (pch != NULL) {
					if (*pch != '#' || *pch != '\n' || strcmp(pch, "\n")) {
						if (sa1.len > 0) {
							// add a space
							if (!stralloc_cats(&sa1, " ")) {
								errno = ENOMEM;
								stralloc_free(&sa0);
								stralloc_free(&sa1);
								free(tmppch);
								free(tmp);
								return errno;
							}
						}
						if (!stralloc_cats(&sa1, pch)) {
							errno = ENOMEM;
							stralloc_free(&sa0);
							stralloc_free(&sa1);
							free(tmppch);
							free(tmp);
							return errno;
						}
					}
					pch = strtok(NULL, "\n");
				}
				if (sa1.len > 0 && strlen(sa1.s) > 0) {
					byte_copy(tmp, sa1.len, sa1.s);
					rwldomains = strdup(tmp);
					if (rwldomains == NULL) {
						errno = ENOMEM;
						stralloc_free(&sa0);
						stralloc_free(&sa1);
						free(tmppch);
						free(tmp);
						return errno;
					}
				}
				free(tmppch);
				free(tmp);
				stralloc_free(&sa1);
			}
			stralloc_free(&sa0);
		}
	}
	if (!rbldomains && rbldomainsfile && *rbldomainsfile) {
		stralloc sa0;
		stralloc_init(&sa0);
		if (!stralloc_copys(&sa0, "")) {
			errno = ENOMEM;
			return errno;
		}
		size_t bufsize = 64;
		if (openreadclose_to_sa(rbldomainsfile, &sa0, bufsize) == 1) {
			if (sa0.len > 0 && strlen(sa0.s) > 0) {
				char *tmp = (char *)malloc(sa0.len+2);
				if (tmp == NULL) {
					errno = ENOMEM;
					str_init(&headermsg);
					stralloc_free(&sa0);
					return errno;
				}
				stralloc sa1;
				stralloc_init(&sa1);
				if (!stralloc_copys(&sa1, "")) {
					errno = ENOMEM;
					stralloc_free(&sa0);
					free(tmp);
					return errno;
				}
				byte_copy(tmp, sa0.len, sa0.s);
				char *pch;
				char *tmppch = strdup(tmp);
				if (tmppch == NULL) {
					errno = ENOMEM;
					stralloc_free(&sa0);
					stralloc_free(&sa1);
					free(tmp);
					return errno;
				}
				pch = strtok(tmppch, "\n");
				while (pch != NULL) {
					if (*pch != '#' || *pch != '\n' || strcmp(pch, "\n")) {
						if (sa1.len > 0) {
							// add a space
							if (!stralloc_cats(&sa1, " ")) {
								errno = ENOMEM;
								stralloc_free(&sa0);
								stralloc_free(&sa1);
								free(tmppch);
								free(tmp);
								return errno;
							}
						}
						if (!stralloc_cats(&sa1, pch)) {
							errno = ENOMEM;
							stralloc_free(&sa0);
							stralloc_free(&sa1);
							free(tmppch);
							free(tmp);
							return errno;
						}
					}
					pch = strtok(NULL, "\n");
				}
				if (sa1.len > 0 && strlen(sa1.s) > 0) {
					byte_copy(tmp, sa1.len, sa1.s);
					rbldomains = strdup(tmp);
					if (rbldomains == NULL) {
						errno = ENOMEM;
						stralloc_free(&sa0);
						stralloc_free(&sa1);
						free(tmppch);
						free(tmp);
						return errno;
					}
				}
				free(tmppch);
				free(tmp);
				stralloc_free(&sa1);
			}
			stralloc_free(&sa0);
		}
	}
	return 0;
}

/* If $RBLSMTPD is set and not empty, first char is - then reply with 553
 * otherwise reply with 451
 * return 1 to block otherwise return 0
 */
int check_rblsmtpd()
{
	const char *x = rblsmtpd;
	if (x) {
		if (!*x) return 0;
		if (*x == '-') {
			resp.number = 553;
			++x;
		} else {
			resp.number = 451;
		}
		resp.message = x;
		return 1;
	}
	return 0;
}

static const response *init(void)
{
	linkproto = getprotoenv(0);
	is_ipv6 = linkproto != 0 && strcasecmp(linkproto, "TCP6") == 0;
	local_ip = getprotoenv("LOCALIP");
	remote_ip = getprotoenv("REMOTEIP");
	relayclient = session_getenv("RELAYCLIENT");
	rejectcheck = session_getenv("DNSBL_CHECK_REJECT");
	defercheck = session_getenv("DNSBL_CHECK_DEFER");
	skipcheck = session_getenv("DNSBL_CHECK_SKIP");
	rblsmtpd = session_getenv("RBLSMTPD");
	logcheck = session_getenv("DNSBL_CHECK_VERBOSE");
	rwldomains = session_getenv("RWLDOMAINS");
	rbldomains = session_getenv("RBLDOMAINS");
	rwldomainsfile = session_getenv("RWLDOMAINSFILE");
	if (!rwldomainsfile) {
		rwldomainsfile = DNSBL_RWLDOMAINS_FILE;
	}
	rbldomainsfile = session_getenv("RBLDOMAINSFILE");
	if (!rbldomainsfile) {
		rbldomainsfile = DNSBL_RBLDOMAINS_FILE;
	}
	thispid = (unsigned long)getpid();
	if (read_dnsbwlfiles() != 0) {
		return &resp_oom;
	}
	return 0;
}

static const response *sender(str* address, str* params)
{
	if (skipcheck) return 0;
	if (!local_ip) {
		local_ip = getprotoenv("LOCALIP");
	}
	if (!local_ip) {
		elogerr(logcheck, 1, "LOCALIP not found thus this plugin skipped\n");
		return 0;
	}
	if (!remote_ip) {
		remote_ip = getprotoenv("REMOTEIP");
	}
	if (!remote_ip) {
		elogerr(logcheck, 1, "REMOTEIP not found thus this plugin skipped\n");
		return 0;
	}
	str_init(&headermsg);
	str_copys(&headermsg, "X-DNSBL-Check-Result: ");
	/* X-DNSBL-Check-Result: SKIPPED: $RELAYCLIENT is set for 1.2.3.4 */
	if (relayclient) {
		elogerr(logcheck, 1, "$RELAYCLIENT is set thus skipped for REMOTEIP %s\n", remote_ip);
		str_cats(&headermsg, "SKIPPED: $RELAYCLIENT is set for ");
		str_cats(&headermsg, remote_ip);
		str_cats(&headermsg, "\n");
		return 0;
	}
	/* X-DNSBL-Check-Result: SKIPPED: authenticated user from 1.2.3.4 */
	if (session_getnum("authenticated", 0)) {
		elogerr(logcheck, 1, "authenticated user thus skipped for REMOTEIP %s\n", remote_ip);
		str_cats(&headermsg, "SKIPPED: authenticated user from ip ");
		str_cats(&headermsg, remote_ip);
		str_cats(&headermsg, "\n");
		return 0;
	}
	/* Need to skip IPv6 for now */
	/* X-DNSBL-Check-Result: SKIPPED: $PROTO is TCP6 or is IPv4 for 1.2.3.4 */
	if (is_ipv6 || strchr(remote_ip, ':') || is_valid_ipv4(remote_ip) != 1) {
		elogerr(logcheck, 1, "$PROTO is TCP6 thus skipped for REMOTEIP %s\n", remote_ip);
		str_cats(&headermsg, "SKIPPED: $PROTO is TCP6 or is non-IPv4 for ");
		str_cats(&headermsg, remote_ip);
		str_cats(&headermsg, "\n");
		return 0;
	}
	/* X-DNSBL-Check-Result: SKIPPED: $RBLSMTPD is set and empty for 1.1.1.1 */
	if (rblsmtpd && rblsmtpd[0] == 0) {
		/* if is set and empty we skip it */
		elogerr(logcheck, 1, "$RBLSMTPD is set and empty thus skipped for REMOTEIP %s\n", remote_ip);
		str_cats(&headermsg, "SKIPPED: $RBLSMTPD is set and empty for ");
		str_cats(&headermsg, remote_ip);
		str_cats(&headermsg, "\n");
		return 0;
	}
	if (check_rblsmtpd()) {
		return &resp;
	}
	/* X-DNSBL-Check-Result: SKIPPED: $RWLDOMAINS and $RBLDOMAINS not set or empty for 1.2.3.4 */
	if ((!rwldomains && !rbldomains) || (!*rwldomains && !*rbldomains)) {
		/* if is set and empty we skip it */
		elogerr(logcheck, 1, "$RWLDOMAINS and $RBLDOMAINS not set or empty thus skipped for REMOTEIP %s\n", remote_ip);
		str_cats(&headermsg, "SKIPPED: $RWLDOMAINS and $RBLDOMAINS not set or empty for ");
		str_cats(&headermsg, remote_ip);
		str_cats(&headermsg, "\n");
		return 0;
	}
	int action = 0;
	if (rejectcheck || defercheck) {
		if (rejectcheck) {
			action = 553;
			resp.number = action;
		} else {
			action = 451;
			resp.number = action;
		}
	}
	char *tmprbl = (char *)malloc(DNS_MAX_DNSBL_LENGTH);
	if (tmprbl == NULL) {
		errno = ENOMEM;
		elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
		str_init(&headermsg);
		return &resp_oom;
	}
	char *delimiters = " ,:\n\t";
	char *ipv4rev = ipv4_reverse((char *)remote_ip);
	if (errno == ENOMEM) {
		free(tmprbl);
		str_init(&headermsg);
		elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
		return &resp_oom;
	}
	char *bouncemsg = (char *)malloc(512);
	if (bouncemsg == NULL) {
		errno = ENOMEM;
		elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
		str_init(&headermsg);
		free(tmprbl);
		return &resp_oom;
	}
	str localmsg;
	str_init(&localmsg);
	int rwl_listed = 0;
	int rbl_listed = 0;
	/* rwldomains */
	if (rwldomains && *rwldomains) {
		char *pch;
		char *tmppch = strdup(rwldomains);
		if (tmppch == NULL) {
			errno = ENOMEM;
			elogerr(1, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
			free(bouncemsg);
			free(tmprbl);
			free(ipv4rev);
			str_free(&localmsg);
			str_init(&headermsg);
			return &resp_oom;
		}
		pch = strtok(tmppch, delimiters);
		while (pch != NULL) {
			if (*pch != '#' && strlen(pch) > 1) {
				sprintf(tmprbl, "%s.%s", ipv4rev, pch);
				elogerr(logcheck, 1, "Checking rwldomain=%s listing for ip=%s query=%s\n", pch, remote_ip, tmprbl);
				struct HostDNSTXT txtres;
				init_HostDNSTXT(&txtres);
				int rc = check_dns_txt(tmprbl, &txtres, MAX_DNSTXT_RECORDS);
				int i = 0;
				if (rc != 0 || errno == ENOMEM) {
					elogerr(1, 1, "(%s:%d) check_dns_txt Out of Memory!\n", __FILE__, __LINE__);
					free(bouncemsg);
					free(tmprbl);
					free(ipv4rev);
					free(tmppch);
					free_HostDNSTXT(&txtres);
					str_free(&localmsg);
					str_init(&headermsg);
					return &resp_oom;
				}
				for (i = 0; i < txtres.size; i++) {
					if (txtres.txtrecord[i]) {
						elogerr(logcheck, "rwldomain=%s ip=%s query=%s result=%s\n", pch, remote_ip, tmprbl, txtres.txtrecord[i]);
						str_cats(&localmsg, "X-DNSWL-Check-Report: ");
						str_cats(&localmsg, "rwldomain=");
						str_cats(&localmsg, pch);
						str_cats(&localmsg, " result=");
						str_cats(&localmsg, txtres.txtrecord[i]);
						str_cats(&localmsg, "\n");
						rwl_listed++;
					}
				}
				free_HostDNSTXT(&txtres);
			}
			pch = strtok(NULL, delimiters);
		}
		free(tmppch);
	}
	/* rbldomains */
	if (rbldomains && *rbldomains) {
		char *pch;
		char *tmppch = strdup(rbldomains);
		if (tmppch == NULL) {
			errno = ENOMEM;
			elogerr(logcheck, 1, "(%s:%d) errno=%d errmsg=%s\n", __FILE__, __LINE__, errno, strerror(errno));
			free(bouncemsg);
			free(tmprbl);
			free(ipv4rev);
			str_free(&localmsg);
			str_init(&headermsg);
			return &resp_oom;
		}
		pch = strtok(tmppch, delimiters);
		while (pch != NULL) {
			if (*pch != '#' && strlen(pch) > 1) {
				sprintf(tmprbl, "%s.%s", ipv4rev, pch);
				elogerr(logcheck, 1, "Checking rbldomain=%s listing for ip=%s query=%s\n", pch, remote_ip, tmprbl);
				struct HostDNSTXT txtres;
				init_HostDNSTXT(&txtres);
				int rc = check_dns_txt(tmprbl, &txtres, MAX_DNSTXT_RECORDS);
				int i = 0;
				if (rc != 0 || errno == ENOMEM) {
					elogerr(logcheck, 1, "(%s:%d) check_dns_txt Out of Memory!\n", __FILE__, __LINE__);
					free(bouncemsg);
					free(tmprbl);
					free(ipv4rev);
					free(tmppch);
					free_HostDNSTXT(&txtres);
					str_free(&localmsg);
					str_init(&headermsg);
					return &resp_oom;
				}
				for (i = 0; i < txtres.size; i++) {
					if (txtres.txtrecord[i]) {
						elogerr(logcheck, 1, "rbldomain=%s ip=%s query=%s result=%s\n", pch, remote_ip, tmprbl, txtres.txtrecord[i]);
						str_cats(&localmsg, "X-DNSBL-Check-Report: ");
						str_cats(&localmsg, "rbldomain=");
						str_cats(&localmsg, pch);
						str_cats(&localmsg, " result=");
						str_cats(&localmsg, txtres.txtrecord[i]);
						str_cats(&localmsg, "\n");
						if (rbl_listed == 0) {
							sprintf(bouncemsg, "Sorry, your ip %s is listed in DNSBL %s and its txt record is %s", remote_ip, pch, txtres.txtrecord[i]);
						}
						rbl_listed++;
					}
				}
				free_HostDNSTXT(&txtres);
			}
			pch = strtok(NULL, delimiters);
		}
		free(tmppch);
	}
	if (rwl_listed > 0) {
		str_cats(&headermsg, "PASSED for ");
		action = 0;
	} else {
		if (rbl_listed > 0) {
			str_cats(&headermsg, "FAILED for ");
		} else {
			str_cats(&headermsg, "PASSED for ");
			action = 0;
		}
	}
	str_cats(&headermsg, remote_ip);
	str_cats(&headermsg, "\n");
	if (localmsg.len > 0) {
		str_cats(&headermsg, localmsg.s);
	}
	if (action > 0 && rbldomains && *rbldomains) {
		if (strlen(bouncemsg) == 0) {
			sprintf(bouncemsg, "Sorry, your ip %s is listed in one or more of the DNSBL: %s", remote_ip, rbldomains);
		}
		str tmpstr;
		str_init(&tmpstr);
		str_copys(&tmpstr, bouncemsg);
		resp.message = tmpstr.s;
		str_free(&tmpstr);
		free(bouncemsg);
		free(ipv4rev);
		free(tmprbl);
		str_free(&localmsg);
		str_init(&headermsg);
		return &resp;
	}
	free(bouncemsg);
	str_free(&localmsg);
	free(tmprbl);
	free(ipv4rev);
	return 0;
	(void)address;
	(void)params;
}

static const response *data_start(int fd)
{
	if (skipcheck) return 0;
	if (headermsg.len > 0) {
		return backend_data_block(headermsg.s, headermsg.len);
	}
	return 0;
	(void)fd;
}

struct plugin plugin = {
	.version = PLUGIN_VERSION,
	.flags = 0,
	.init = init,
	.sender = sender,
	.data_start = data_start,
};